package flink.zanxiangnet.ad.monitoring.maxcompute.sink;

import com.aliyun.odps.Odps;
import com.aliyun.odps.Table;
import com.aliyun.odps.account.Account;
import com.aliyun.odps.account.AliyunAccount;
import com.aliyun.odps.tunnel.TableTunnel;
import flink.zanxiangnet.ad.monitoring.maxcompute.MaxComputeLog;
import flink.zanxiangnet.ad.monitoring.maxcompute.bean.BeanUtil;
import flink.zanxiangnet.ad.monitoring.maxcompute.bean.annotation.MaxComputeTable;
import flink.zanxiangnet.ad.monitoring.pojo.properties.ApplicationProperties;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.functions.sink.RichSinkFunction;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.Map;

public class TunnelBatchSink<IN> extends RichSinkFunction<IN> {
    private static final Logger log = LoggerFactory.getLogger(TunnelBatchSink.class);

    // 对象锁，防止MaxCompute的 Tunnel对象多次初始化
    private static final Object DUMMY_LOCK = new Object();

    private final Class<IN> clazz;
    // 缓存刷新的间隔时间
    private final Long bufferRefreshTime;
    // 缓存的最大数据量
    private final Long maxBufferCount;
    // 可能同时存在的分区数
    private final Integer partitionCount;

    private volatile transient TunnelBatchSinkBuffer<IN> sinkBuffer;

    public TunnelBatchSink(Class<IN> clazz, Long bufferRefreshTime, Long maxBufferCount, Integer partitionCount) {
        this.clazz = clazz;
        this.bufferRefreshTime = bufferRefreshTime;
        this.maxBufferCount = maxBufferCount;
        this.partitionCount = partitionCount;
    }

    @Override
    public void open(Configuration config) {
        if (sinkBuffer == null) {
            synchronized (DUMMY_LOCK) {
                if (sinkBuffer == null) {
                    Map<String, String> params = getRuntimeContext()
                            .getExecutionConfig()
                            .getGlobalJobParameters()
                            .toMap();
                    MaxComputeTable tableAnnotation = clazz.getAnnotation(MaxComputeTable.class);

                    Account account = new AliyunAccount(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ID),
                            params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_KEY));
                    Odps odps = new Odps(account);
                    odps.getRestClient().setRetryLogger(new MaxComputeLog());
                    odps.setEndpoint(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ENDPOINT));
                    odps.setDefaultProject(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_PROJECT_NAME));
                    TableTunnel tunnel = new TableTunnel(odps);
                    tunnel.setEndpoint(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_TUNNEL_ENDPOINT));
                    Table table = odps.tables().get(tableAnnotation.value());
                    List<BeanUtil.FieldInfo> fieldInfoList = BeanUtil.parseBeanField(clazz);
                    sinkBuffer = new TunnelBatchSinkBuffer<>(bufferRefreshTime,
                            maxBufferCount,
                            partitionCount,
                            tunnel,
                            params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_PROJECT_NAME),
                            table,
                            fieldInfoList);
                }
            }
        }
    }

    /**
     * 将值写入到 Sink。每个值都会调用此函数
     *
     * @param value
     * @param context
     */
    @Override
    public void invoke(IN value, Context context) {
        sinkBuffer.put(value);
    }

    @Override
    public void close() throws Exception {
        if (sinkBuffer != null) {
            synchronized (DUMMY_LOCK) {
                sinkBuffer.close();
            }
        }

        super.close();
    }
}
