package flink.zanxiangnet.ad.monitoring.maxcompute.sink;

import com.aliyun.odps.PartitionSpec;
import com.aliyun.odps.Table;
import com.aliyun.odps.tunnel.TableTunnel;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import flink.zanxiangnet.ad.monitoring.maxcompute.bean.BeanUtil;
import flink.zanxiangnet.ad.monitoring.util.JsonUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.util.CollectionUtils;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.*;
import java.util.concurrent.*;
import java.util.stream.Collectors;

/**
 * 缓存要写入 MaxCompute的数据
 */
public class TunnelBatchSinkBuffer<T> {
    private static final Logger log = LoggerFactory.getLogger(TunnelBatchSinkBuffer.class);

    private static final String PART = "PART";
    private static final String NO_PART = "NO_PART";

    private final TableTunnel tunnel;
    private final String projectName;
    private final Table tableInfo;

    private final Long buffRefreshTime;
    private final Long maxBuffer;

    private final ConcurrentHashMap<String, List<T>> partitionData;
    private final Map<String, Method> partitionFieldMethods;
    private final List<BeanUtil.FieldInfo> fieldInfoList;

    private boolean isRun = false;
    private Thread coreThread;
    private final ExecutorService threadPool;

    public TunnelBatchSinkBuffer(
            Long buffRefreshTime,
            Long maxBuffer,
            Integer partitionCount,
            TableTunnel tunnel,
            String projectName,
            Table tableInfo,
            List<BeanUtil.FieldInfo> fieldInfoList
    ) {
        this.tunnel = tunnel;
        this.projectName = projectName;
        this.tableInfo = tableInfo;
        this.buffRefreshTime = buffRefreshTime;
        this.maxBuffer = maxBuffer;
        this.fieldInfoList = fieldInfoList;
        this.partitionData = new ConcurrentHashMap<>(partitionCount);

        this.partitionFieldMethods = fieldInfoList.stream().filter(BeanUtil.FieldInfo::isUsePartitioned).collect(Collectors.toMap(BeanUtil.FieldInfo::getColumnName, BeanUtil.FieldInfo::getGetMethod));
        threadPool = new ThreadPoolExecutor(
                partitionCount,
                partitionCount,
                0L,
                TimeUnit.MILLISECONDS,
                new LinkedBlockingQueue<>(),
                new ThreadFactoryBuilder()
                        .setNameFormat("maxcompute-writer-%d").build());
        start();
    }

    /**
     * 写入数据
     *
     * @param t 数据
     */
    public void put(T t) {
        StringBuilder partition;
        if (CollectionUtils.isEmpty(partitionFieldMethods)) {
            partition = new StringBuilder(NO_PART);
        } else {
            partition = new StringBuilder(PART);
            for (Map.Entry<String, Method> entry : partitionFieldMethods.entrySet()) {
                partition.append(entry.getKey()).append("=");
                try {
                    partition.append(entry.getValue().invoke(t));
                } catch (InvocationTargetException | IllegalAccessException e) {
                    // 获取分区字段的值失败
                    log.error(e.getMessage(), e);
                    throw new RuntimeException("Failed get partition field value!");
                }
                partition.append(",");
            }
            partition = new StringBuilder(partition.substring(0, partition.length() - 1));
        }
        List<T> list = partitionData.computeIfAbsent(partition.toString(), key -> new ArrayList<>());
        list.add(t);
    }

    public void close() {
        isRun = false;
        threadPool.shutdown();
        coreThread.interrupt();
    }

    private void start() {
        coreThread = new Thread(() -> {
            Map<String, Long> lastSaveTime = new HashMap<>();
            // 开启一个定时轮询缓存数据的线程
            while (isRun) {
                try {
                    if (CollectionUtils.isEmpty(partitionData)) {
                        // 没有数据，等 5s
                        try {
                            Thread.sleep(5000L);
                        } catch (InterruptedException e) {
                            Thread.sleep(5000L);
                        }
                        continue;
                    }
                    long now = System.currentTimeMillis();
                    List<String> partitionKeys = new ArrayList<>(partitionData.size());
                    for (Map.Entry<String, List<T>> entry : partitionData.entrySet()) {
                        Long lastSaveTimeOfKey = lastSaveTime.computeIfAbsent(entry.getKey(), key -> now);
                        boolean mustSave = now - lastSaveTimeOfKey > buffRefreshTime;
                        if (mustSave || entry.getValue().size() > maxBuffer) {
                            partitionKeys.add(entry.getKey());
                        }
                    }
                    if (CollectionUtils.isEmpty(partitionKeys)) {
                        Thread.sleep(5000L);
                        continue;
                    }

                    for (String partitionKey : partitionKeys) {
                        List<T> dataList = partitionData.remove(partitionKey);
                        lastSaveTime.remove(partitionKey);

                        PartitionSpec part = null;
                        if (partitionKey.startsWith(PART)) {
                            part = new PartitionSpec(partitionKey.substring(PART.length()));
                            // 尽然不会自己建分区！！！
                            // 创建单个分区大概 1~5秒（大部分在 2、3秒）
                            tableInfo.createPartition(part, true);
                        }
                        threadPool.submit(new TunnelBatchWriter<>(tunnel, projectName, tableInfo.getName(), part, dataList, fieldInfoList));
                    }
                } catch (Exception e) {
                    log.error(e.getMessage(), e);
                }
            }
        });
        isRun = true;
        coreThread.start();
    }
}
