package flink.zanxiangnet.ad.monitoring.process;

import com.aliyun.odps.Instance;
import com.aliyun.odps.Odps;
import com.aliyun.odps.account.Account;
import com.aliyun.odps.account.AliyunAccount;
import com.aliyun.odps.data.Record;
import com.aliyun.odps.task.SQLTask;
import flink.zanxiangnet.ad.monitoring.maxcompute.MaxComputeLog;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdDataOfMinuteODS;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdStatOfDayDWD;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdStatOfMinuteDWD;
import flink.zanxiangnet.ad.monitoring.pojo.properties.ApplicationProperties;
import flink.zanxiangnet.ad.monitoring.util.DateUtil;
import flink.zanxiangnet.ad.monitoring.util.JsonUtil;
import org.apache.flink.api.common.state.MapState;
import org.apache.flink.api.common.state.MapStateDescriptor;
import org.apache.flink.api.common.state.ValueState;
import org.apache.flink.api.common.state.ValueStateDescriptor;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.functions.windowing.ProcessWindowFunction;
import org.apache.flink.streaming.api.windowing.windows.TimeWindow;
import org.apache.flink.util.Collector;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

public class AdMinuteDWDProcess extends ProcessWindowFunction<AdDataOfMinuteODS, AdStatOfMinuteDWD, Long, TimeWindow> {
    private static final DateTimeFormatter formatForLastReduceKey = DateTimeFormatter.ofPattern("yyyy-MM-dd_HH:mm");
    private Odps odps;

    // 历史的天数据
    private ValueState<AdStatOfDayDWD> historyDayState;
    // 上次查询的天数据
    private ValueState<String> lastQueryDayState;
    // 之前聚合的昨天的数据
    private MapState<String, AdStatOfMinuteDWD> yesterdayMinuteState;
    // 前 5分钟聚合的数据
    private MapState<String, AdStatOfMinuteDWD> lastReduceState;

    @Override
    public void open(Configuration conf) {
        Map<String, String> params = getRuntimeContext()
                .getExecutionConfig()
                .getGlobalJobParameters()
                .toMap();
        Account account = new AliyunAccount(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ID),
                params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_KEY));
        odps = new Odps(account);
        odps.getRestClient().setRetryLogger(new MaxComputeLog());
        odps.setEndpoint(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ENDPOINT));
        odps.setDefaultProject(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_PROJECT_NAME));

        historyDayState = getRuntimeContext().getState(new ValueStateDescriptor<>("historyDayState", AdStatOfDayDWD.class));
        lastQueryDayState = getRuntimeContext().getState(new ValueStateDescriptor<>("lastQueryDayState", String.class));
        yesterdayMinuteState = getRuntimeContext().getMapState(new MapStateDescriptor<>("yesterdayMinuteState", String.class, AdStatOfMinuteDWD.class));
        lastReduceState = getRuntimeContext().getMapState(new MapStateDescriptor<>("lastReduceState", String.class, AdStatOfMinuteDWD.class));
    }

    @Override
    public void process(Long elementCount, ProcessWindowFunction<AdDataOfMinuteODS, AdStatOfMinuteDWD, Long, TimeWindow>.Context context,
                        Iterable<AdDataOfMinuteODS> iterable, Collector<AdStatOfMinuteDWD> collector) throws Exception {
        long beginTime = context.window().getStart();
        LocalDateTime beginDateTime = DateUtil.milliToLocalDateTime(beginTime);
        LocalDate beginDate = beginDateTime.toLocalDate();
        String statDay = DateUtil.formatLocalDate(beginDate);
        int hour = beginDateTime.getHour();
        long now = System.currentTimeMillis();
        AdDataOfMinuteODS statODS = null;
        List<AdDataOfMinuteODS> adDataOfMinuteODSList = new ArrayList<>(24);
        for (AdDataOfMinuteODS adDataOfMinuteODS : iterable) {
            adDataOfMinuteODSList.add(adDataOfMinuteODS);
            if (adDataOfMinuteODS.getHour() != hour) {
                continue;
            }
            statODS = adDataOfMinuteODS;
            break;
        }
        System.out.println("windowCount：" + adDataOfMinuteODSList.size());
        if (statODS == null) {
            return;
        }

        Long adId = statODS.getAdId();

        // 昨天聚合的数据
        AdStatOfMinuteDWD yesterdayMinuteDWD = yesterdayMinuteState.get(DateUtil.formatLocalDate(beginDate.minusDays(1L)));
        // 之前的数据
        String lastQueryDay = lastQueryDayState.value();
        if (lastQueryDay == null || !lastQueryDay.equals(statDay)) {
            // 往前找 30天
            LocalDate endDay = beginDate.minusDays(2L), beginDay = endDay.minusDays(30L);
            Instance instance = SQLTask.run(odps, "SELECT * FROM ad_stat_of_day_dwd WHERE stat_day >= \"" + DateUtil.formatLocalDate(beginDay) + "\" AND stat_day <= \"" + endDay + "\" AND ad_id = " + adId + ";");
            instance.waitForSuccess();
            List<Record> records = SQLTask.getResult(instance);
            List<AdStatOfDayDWD> historyDayData = records.stream().map(AdStatOfDayDWD::byMaxCompute).sorted((val1, val2) -> {
                if (val1.getStatDay().equals(val2.getStatDay())) {
                    return new Long(val1.getCreateTime().getTime() - val2.getCreateTime().getTime()).intValue();
                }
                return DateUtil.parseLocalDate(val1.getStatDay()).compareTo(DateUtil.parseLocalDate(val2.getStatDay()));
            }).collect(Collectors.toList());
            if (!historyDayData.isEmpty()) {
                historyDayState.update(historyDayData.get(historyDayData.size() - 1));
            }
            lastQueryDayState.update(statDay);
        }
        AdStatOfDayDWD beforeYesterdayDayDWD = historyDayState.value();

        AdStatOfMinuteDWD lastReduceData = lastReduceState.get(beginDateTime.minusMinutes(5L).format(formatForLastReduceKey));
        // 聚合当天的全部数据
        AdStatOfMinuteDWD newAdStat = AdStatOfMinuteDWD.reduce(beforeYesterdayDayDWD, yesterdayMinuteDWD, adDataOfMinuteODSList, statODS, lastReduceData, now);
        collector.collect(newAdStat);
        System.out.println("输出：" + JsonUtil.toString(newAdStat));
        lastReduceState.put(beginDateTime.format(formatForLastReduceKey), newAdStat);
        yesterdayMinuteState.put(statDay, newAdStat);
        
        for (String key : lastReduceState.keys()) {
            if (key.equals(beginDateTime.format(formatForLastReduceKey))
                    || key.equals(beginDateTime.minusMinutes(5L).format(formatForLastReduceKey))
                    || key.equals(beginDateTime.minusMinutes(10L).format(formatForLastReduceKey))) {
                continue;
            }
            lastReduceState.remove(key);
        }
        for (String key : yesterdayMinuteState.keys()) {
            if (key.equals(DateUtil.formatLocalDate(beginDate))
                    || key.equals(DateUtil.formatLocalDate(beginDate.minusDays(1L)))
                    || key.equals(DateUtil.formatLocalDate(beginDate.minusDays(2L)))) {
                continue;
            }
            yesterdayMinuteState.remove(key);
        }
    }

    @Override
    public void clear(ProcessWindowFunction<AdDataOfMinuteODS, AdStatOfMinuteDWD, Long, TimeWindow>.Context context) throws Exception {
        System.out.println("窗口关闭");
    }
}
