package flink.zanxiangnet.ad.monitoring.trigger;

import org.apache.flink.api.common.state.ValueState;
import org.apache.flink.api.common.state.ValueStateDescriptor;
import org.apache.flink.streaming.api.windowing.triggers.Trigger;
import org.apache.flink.streaming.api.windowing.triggers.TriggerResult;
import org.apache.flink.streaming.api.windowing.windows.Window;

/**
 * 定时 && 定量触发
 *
 * @param <T>
 * @param <W>
 */
public class TimerCountTrigger<T, W extends Window> extends Trigger<T, W> {
    private static final String STATE_NAME_WINDOW_TIME = "_windowTime";
    private static final String STATE_NAME_WINDOW_COUNT = "_windowCount";

    // 缓存刷新的间隔时间
    private final Long bufferRefreshTime;
    // 缓存的最大数据量
    private final Long maxBufferCount;
    // 缓存架构
    private final String triggerName;

    public TimerCountTrigger(String triggerName, Long maxBufferCount, Long bufferRefreshTime) {
        this.triggerName = triggerName;
        this.maxBufferCount = maxBufferCount;
        this.bufferRefreshTime = bufferRefreshTime;
    }

    /**
     * 每次元素进入时调用
     *
     * @param element        元素
     * @param time           元素进入的时间（这个时间有问题）
     * @param window         元素被添加到的窗口对象
     * @param triggerContext trigger上下文
     * @return
     * @throws Exception
     */
    @Override
    public TriggerResult onElement(T element, long time, W window, TriggerContext triggerContext) throws Exception {
        ValueState<Long> windowTimeState = triggerContext.getPartitionedState(new ValueStateDescriptor<>(triggerName + STATE_NAME_WINDOW_TIME, Long.class));
        ValueState<Long> windowCountState = triggerContext.getPartitionedState(new ValueStateDescriptor<>(triggerName + STATE_NAME_WINDOW_COUNT, Long.class));
        long now = System.currentTimeMillis();
        Long windowCount = windowCountState.value() == null ? 1L : windowCountState.value() + 1;
        Long windowTime = windowTimeState.value() == null ? now : windowTimeState.value();
        if (windowCount == 1 || windowTime == now) {
            // 注册一个定时器，到时间了去触发 ProcessTime
            triggerContext.registerProcessingTimeTimer(triggerContext.getCurrentProcessingTime() + bufferRefreshTime);
        }

        if (windowCount >= maxBufferCount || (now - windowTime) >= bufferRefreshTime) {
            windowCountState.update(0L);
            windowTimeState.update(now);
            return TriggerResult.FIRE_AND_PURGE;
        }

        windowCountState.update(windowCount);
        if (windowTimeState.value() == null) {
            windowTimeState.update(now);
        }

        return TriggerResult.CONTINUE;
    }

    /**
     * trigger上下文设置的 processTimer触发时调用
     *
     * @param time           计时器触发的时间戳
     * @param window
     * @param triggerContext
     * @return
     * @throws Exception
     */
    @Override
    public TriggerResult onProcessingTime(long time, W window, TriggerContext triggerContext) throws Exception {
        ValueState<Long> windowTimeState = triggerContext.getPartitionedState(new ValueStateDescriptor<>(triggerName + STATE_NAME_WINDOW_TIME, Long.class));
        // long now = System.currentTimeMillis();
        long now = time;
        Long windowTime = windowTimeState.value();
        if (now - windowTime >= maxBufferCount) {
            triggerContext.getPartitionedState(new ValueStateDescriptor<>(triggerName + STATE_NAME_WINDOW_TIME, Long.class)).update(0L);
            windowTimeState.update(now);
            return TriggerResult.FIRE_AND_PURGE;
        }
        return TriggerResult.CONTINUE;
    }

    /**
     * trigger上下文设置的 eventTimer触发时调用
     *
     * @param time
     * @param window
     * @param triggerContext
     * @return
     * @throws Exception
     */
    @Override
    public TriggerResult onEventTime(long time, W window, TriggerContext triggerContext) throws Exception {
        return TriggerResult.CONTINUE;
    }

    /**
     * 窗口被清除时调用
     *
     * @param w
     * @param triggerContext
     * @throws Exception
     */
    @Override
    public void clear(W w, TriggerContext triggerContext) throws Exception {
        ValueState<Long> windowTime = triggerContext.getPartitionedState(new ValueStateDescriptor<>(triggerName + STATE_NAME_WINDOW_TIME, Long.class));
        ValueState<Long> windowCount = triggerContext.getPartitionedState(new ValueStateDescriptor<>(triggerName + STATE_NAME_WINDOW_COUNT, Long.class));
        windowTime.clear();
        windowCount.clear();
    }
}
